﻿using System;
using System.Collections.Generic;
using System.Data;
using System.Drawing;
using System.Windows.Forms;
using System.Security.Cryptography;
using System.Text;
using System.Net;
using System.Net.Http;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using System.IO;
using System.Configuration;
using System.Media;
using System.Reflection;

using RestSharp;
using Newtonsoft.Json;



namespace BlendoStreamSuite
{
    public partial class Form1 : Form
    {
        private string discordURL;
        private string twitchToken, twitchChannelName;
        private string twitterConsumerKey, twitterConsumerSecretKey, twitterAccessToken, twitterAccessSecretToken;

        private const string KEYSFILE = "keys.txt";

        private bool hasPlayedBeep = false;

        System.Threading.Thread t;

        public Form1()
        {
            InitializeComponent();

            this.FormClosing += new System.Windows.Forms.FormClosingEventHandler(this.Form1_FormClosing);
            this.Resize += new System.EventHandler(Form1_Resize);



            //Retrieve the keys.
            ReadKeys();


            //Address config file crash.
            string currentFolder = Path.GetDirectoryName(Assembly.GetExecutingAssembly().Location);
            string[] configFiles = Directory.GetFiles(currentFolder, "*.exe.config");

            if (configFiles.Length <= 0)
            {
                MessageBox.Show("Missing .config file!");
                Environment.Exit(0);
                return;
            }


            Properties.Settings.Default.Upgrade();
            
            

            UpdateLastPressedTime();


            DoTweet(false);
            DoDiscordBot(false);

            DoTwitchTitle();

            

            DoCountdown();

            Shown += Form1_Shown;
        }

        private void ReadKeys()
        {
            if (!File.Exists(KEYSFILE))
            {
                MessageBox.Show("CATASTROPHIC ERROR: Couldn't find keys.txt");
                SetKeysEmpty();
                return;
            }

            string rawContents = GetFileContents(KEYSFILE);

            rawContents = rawContents.Trim();

            if (string.IsNullOrWhiteSpace(rawContents))
            {
                MessageBox.Show("CATASTROPHIC ERROR: keys.txt is empty!");
                SetKeysEmpty();
                return;
            }

            string[] lines = rawContents.Split(new string[] { "\n", "\r" }, StringSplitOptions.RemoveEmptyEntries);

            List<string> valueLines = new List<string>();

            for (int i = 0; i < lines.Length; i++)
            {
                lines[i] = lines[i].Trim();

                if (lines[i].StartsWith("//"))
                    continue;

                valueLines.Add(lines[i]);
            }

            if (valueLines.Count < 7)
            {
                MessageBox.Show(string.Format("CATASTROPHIC ERROR: invalid amount of keys in keys.txt\n\nExpected 7 keys, found {0}.", valueLines.Count));
                SetKeysEmpty();
                return;
            }

            try
            {
                discordURL = valueLines[0];
                twitchToken = valueLines[1];
                twitchChannelName = valueLines[2];
                twitterConsumerKey = valueLines[3];
                twitterConsumerSecretKey = valueLines[4];
                twitterAccessToken = valueLines[5];
                twitterAccessSecretToken = valueLines[6];
            }
            catch (Exception e)
            {
                MessageBox.Show(string.Format("CATASTROPHIC ERROR: there was an error parsing keys.txt\n\n{0}", e.Message));
                SetKeysEmpty();
                return;
            }

            //Read all the keys, done.
        }

        private void SetKeysEmpty()
        {
            discordURL = "0";
            twitchToken = "0";
            twitchChannelName = "0";
            twitterConsumerKey = "0";
            twitterConsumerSecretKey = "0";
            twitterAccessToken = "0";
            twitterAccessSecretToken = "0";
        }

        public string GetFileContents(string filepath)
        {
            string output = string.Empty;

            try
            {
                using (FileStream stream = File.Open(filepath, FileMode.Open, FileAccess.Read))
                {
                    using (StreamReader reader = new StreamReader(stream))
                    {
                        //dump file contents into a string.
                        output = reader.ReadToEnd();
                    }
                }
            }
            catch
            {
                return null;
            }

            return output;
        }

        private void UpdateLastPressedTime()
        {
            DateTime lastPostTime = Properties.Settings.Default.lastPostTime;

            DateTime timeNow = DateTime.Now;
            TimeSpan differential = timeNow.Subtract(lastPostTime);

            string lastPostTimeStr = lastPostTime.ToString("hh:mm tt");
            string lastPostDate = lastPostTime.ToString("MM/dd/yyyy");

            if (lastPostTime == DateTime.MinValue)
                lastPostLabel.Text = "Last post: none!";
            else if (differential.Days > 365)
                lastPostLabel.Text = string.Format("Last post: more than a year ago. ({0} {1})", lastPostDate, lastPostTimeStr);
            else if (differential.Days > 0)
                lastPostLabel.Text = string.Format("Last post: {0} days ago. ({1} {2})", differential.Days, lastPostDate, lastPostTimeStr);
            else if (differential.Hours > 0)
                lastPostLabel.Text = string.Format("Last post: {0} hours ago. ({1})", differential.Hours, lastPostTimeStr);
            else if (differential.TotalMinutes > 2)
                lastPostLabel.Text = string.Format("Last post: {0} minutes ago. ({1})", (int)differential.TotalMinutes, lastPostTimeStr); //More than 2 minutes, less than an hour.
            else
                lastPostLabel.Text = string.Format("Last post: just now! ({0})", lastPostTimeStr); //under two minutes = "just now"
        }

        private void Form1_Resize(object sender, System.EventArgs e)
        {
            /*
            if (this.WindowState == FormWindowState.Minimized)
            {
                Hide();
                notifyIcon1.Visible = true;
            }*/
        }

        private void Form1_Shown(object sender, EventArgs e)
        {
            t = new System.Threading.Thread(DoThisAllTheTime);
            t.Start();
        }

        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            t.Abort();
            base.OnClosing(e);
        }

        private void DoTwitchTitle()
        {
            string todaysDateString = string.Format(Properties.Settings.Default.TwitchTitle, DateTime.Now.ToString("MMM"), DateTime.Today.Day, DateTime.Today.Year);
            twitchTitleLabel.Text = todaysDateString;

           

            var client = new RestClient(string.Format("https://api.twitch.tv/kraken/channels/{0}", twitchChannelName));
            var request = new RestRequest(Method.PUT);
            request.AddHeader("cache-control", "no-cache");
            request.AddHeader("content-type", "application/json");
            request.AddHeader("authorization", "OAuth " + twitchToken);
            request.AddHeader("accept", "application/vnd.twitchtv.v3+json");
            request.AddParameter("application/json", "{\"channel\":{\"status\":\"" + todaysDateString + "\"}}", ParameterType.RequestBody);

            IRestResponse response = client.Execute(request);

            if (response.IsSuccessful)
            {                
                SetLabelSuccess(twitchLabelSuccess, true);
            }
            else
            {
                //MessageBox.Show(string.Format("Failed to set Twitch stream title.\n\n{0}", response.ErrorMessage));
                SetLabelSuccess(twitchLabelSuccess, false);
            }
        }


        private void DoTweet(bool send)
        {
            string tweetExample  = InitializeTweetMessage(Properties.Settings.Default.TweetMessage);
            exampleTweetLabel.Text = tweetExample;

            if (!send)
                return;
            
            if (string.IsNullOrWhiteSpace(tweetExample))
                return;

            SendTweetAsync(tweetExample, Properties.Settings.Default.TwitterReplyID, twitterConsumerKey, twitterConsumerSecretKey, twitterAccessToken, twitterAccessSecretToken);
        }



        private string InitializeTweetMessage(string baseString)
        {
            int[] startValue = GetStartTimeValue();

            if (startValue == null)
                return null;

            DateTime startTime = DateTime.Now.Date.Add(new TimeSpan(startValue[0], startValue[1], 0));
            DateTime timeNow = DateTime.Now;
            TimeSpan differential = startTime.Subtract(timeNow);

            int minutesDifferential = Math.Max( differential.Minutes, 2);

            string tweetExample = string.Format(baseString, minutesDifferential);
            
            return tweetExample;
        }

        private void DoDiscordBot(bool send)
        {
            HttpClient client = new HttpClient();

            string message = InitializeTweetMessage(Properties.Settings.Default.discordMessage);
            discordLabel.Text = message;

            if (!send)
                return;
            

            DiscordMsg discordMessage = new DiscordMsg(message);
            var discordMessageJSON = JsonConvert.SerializeObject(discordMessage, Formatting.Indented);
            string results = string.Empty;

            try
            {
                var resp = client.PostAsync(discordURL, new StringContent(discordMessageJSON, Encoding.UTF8, "application/json")).Result;
                results = resp.ToString();
            }
            catch
            {
            }

            if (results.ToString().Contains("NO CONTENT") || results.ToString().Contains("204"))
            {
                SetLabelSuccess(discordSuccess, true);
            }
            else
            {
                SetLabelSuccess(discordSuccess, false);
            }
        }

        private int[] GetStartTimeValue()
        {
            string startTimeRaw = Properties.Settings.Default.StreamStartTime;

            int colonIndex = startTimeRaw.IndexOf(':');

            if (colonIndex < 0)
            {
                MessageBox.Show(string.Format("Invalid stream start time: {0}\n\nStream start time examples:\n10:00\n16:30", startTimeRaw));
                return null;
            }


            try
            {
                int hourNumber = int.Parse(startTimeRaw.Substring(0, colonIndex));
                int minuteNumber = int.Parse(startTimeRaw.Substring(colonIndex + 1, startTimeRaw.Length - colonIndex - 1));
                return new int[2] { hourNumber, minuteNumber };
            }
            catch
            {
                return null;
            }
        }

        private void DoCountdown()
        {
            string todaysDateString = string.Format(Properties.Settings.Default.StreamDate, DateTime.Now.ToString("MMMM"), DateTime.Today.Day, DateTime.Today.Year);
            streamDate.Text = todaysDateString;

            WriteTextFile(Properties.Settings.Default.dateTextfile, todaysDateString);
        }

        public void DoThisAllTheTime()
        {
            int[] startValue = GetStartTimeValue();

            if (startValue == null)
                return;

            DateTime startTime = DateTime.Now.Date.Add(new TimeSpan(startValue[0], startValue[1], 0));

            while (true)
            {

                DateTime timeNow = DateTime.Now;
                TimeSpan differential = startTime.Subtract(timeNow);

                string countdownText = string.Empty;

                if (differential.TotalMinutes <= 0 && differential.TotalSeconds <= 0 && differential.TotalHours <= 0)
                {
                    //done.
                    countdownText = Properties.Settings.Default.CountdownDone;

                    if (!hasPlayedBeep)
                    {
                        hasPlayedBeep = true;
                        Console.Beep(300, 150);
                    }
                }
                else
                {
                    countdownText = string.Format(Properties.Settings.Default.CountdownText, (differential.Hours * 60) + differential.Minutes, differential.Seconds.ToString("D2"));
                }


                if (string.IsNullOrWhiteSpace(countdownText))
                    continue;
                                

                //you need to use Invoke because the new thread can't access the UI elements directly
                MethodInvoker mi = delegate () { streamCountdown.Text = countdownText; };
                this.Invoke(mi);



                using (FileStream fileStream = File.Open(Properties.Settings.Default.countdownTextfile, FileMode.OpenOrCreate, System.IO.FileAccess.Write, FileShare.Read))
                {
                    using (StreamWriter streamWriter = new StreamWriter(fileStream))
                    {
                        streamWriter.WriteLine(countdownText);
                    }
                }


                if (differential.TotalMinutes <= 0 && differential.TotalSeconds <= 0 && differential.TotalHours <= 0)
                {
                    //streamWriter.Close();
                    t.Abort();
                    return;
                }
            }
        }

        private void WriteTextFile(string filepath, string text)
        {
            using (System.IO.StreamWriter file = new System.IO.StreamWriter(filepath, false))
            {
                file.WriteLine(text);
            }
        }





        private async void SendTweetAsync(string message, string threadReplyID, string consumerKey, string consumerSecretKey, string accessToken, string accessSecretToken)
        {
            var json = await SendTweet(message, threadReplyID, consumerKey, consumerSecretKey, accessToken, accessSecretToken);

            if (json.Contains("errors"))
            {
                SetLabelSuccess(tweetSuccess, false);
            }
            else
            {
                SetLabelSuccess(tweetSuccess, true);                
            }
        }

        private void SetLabelSuccess(Label label, bool value)
        {
            if (value)
            {
                label.ForeColor = Color.MediumSeaGreen;
                label.Text = "POSTED!";
            }
            else
            {
                label.ForeColor = Color.Red;
                label.Text = "FAILED.";
            }
        }


        private static async Task<string> SendTweet(string message, string threadReplyID, string consumerKey, string consumerSecretKey, string accessToken, string accessSecretToken)
        {

            var twitter = new TwitterApi(consumerKey,
                consumerSecretKey,
                accessToken,
                accessSecretToken);

            var response = await twitter.TweetReply(message, threadReplyID);

            return response;
        }

        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Environment.Exit(0);
        }

        private void configureToolStripMenuItem_Click(object sender, EventArgs e)
        {
            configscreen newConfigScreen = new configscreen();
            newConfigScreen.FormClosing += new FormClosingEventHandler(configscreen_closing);

            Enabled = false; //hack to make it look disabled.
            newConfigScreen.ShowDialog();
            Enabled = true;  //hack to make it look disabled.
        }

        void configscreen_closing(object sender, FormClosingEventArgs e)
        {
            //when the editscreen closes, then update this list.
            string tweetMsg = InitializeTweetMessage(Properties.Settings.Default.TweetMessage);
            exampleTweetLabel.Text = tweetMsg;

            string discordMsg = InitializeTweetMessage(Properties.Settings.Default.discordMessage);
            discordLabel.Text = discordMsg;

            string todaysDateString = string.Format(Properties.Settings.Default.TwitchTitle, DateTime.Now.ToString("MMM"), DateTime.Today.Day, DateTime.Today.Year);
            twitchTitleLabel.Text = todaysDateString;
        }

        private void notifyIcon1_MouseDoubleClick(object sender, MouseEventArgs e)
        {
            /*
            Show();
            this.WindowState = FormWindowState.Normal;
            notifyIcon1.Visible = false;*/
        }

        private void button1_Click(object sender, EventArgs e)
        {
            //Post the twitter + discord messages.
            DoTweet(true);
            DoDiscordBot(true);

            Properties.Settings.Default.lastPostTime = DateTime.Now;
            Properties.Settings.Default.Save();

            UpdateLastPressedTime();
        }
    }

    class DiscordMsg
    {
        [JsonProperty("content")]
        public string content { get; set; }

        public DiscordMsg(string message)
        {
            content = message;
        }
    }



    //From: https://blog.dantup.com/2016/07/simplest-csharp-code-to-post-a-tweet-using-oauth
    class TwitterApi
    {
        const string TwitterApiBaseUrl = "https://api.twitter.com/1.1/";
        readonly string consumerKey, consumerKeySecret, accessToken, accessTokenSecret;
        readonly HMACSHA1 sigHasher;
        readonly DateTime epochUtc = new DateTime(1970, 1, 1, 0, 0, 0, DateTimeKind.Utc);

        /// <summary>
        /// Creates an object for sending tweets to Twitter using Single-user OAuth.
        /// 
        /// Get your access keys by creating an app at apps.twitter.com then visiting the
        /// "Keys and Access Tokens" section for your app. They can be found under the
        /// "Your Access Token" heading.
        /// </summary>
        public TwitterApi(string consumerKey, string consumerKeySecret, string accessToken, string accessTokenSecret)
        {
            this.consumerKey = consumerKey;
            this.consumerKeySecret = consumerKeySecret;
            this.accessToken = accessToken;
            this.accessTokenSecret = accessTokenSecret;

            sigHasher = new HMACSHA1(new ASCIIEncoding().GetBytes(string.Format("{0}&{1}", consumerKeySecret, accessTokenSecret)));
        }

        /// <summary>
        /// Sends a tweet with the supplied text and returns the response from the Twitter API.
        /// </summary>
        public Task<string> Tweet(string text)
        {
            var data = new Dictionary<string, string> {
                { "status", text },
                { "trim_user", "1" }
            };

            return SendRequest("statuses/update.json", data);
        }

        public Task<string> TweetReply(string text, string threadID)
        {
            var data = new Dictionary<string, string> {
                { "status", text },
                { "trim_user", "1" },
                { "in_reply_to_status_id", threadID }
            };

            return SendRequest("statuses/update.json", data);
        }

        Task<string> SendRequest(string url, Dictionary<string, string> data)
        {
            var fullUrl = TwitterApiBaseUrl + url;

            // Timestamps are in seconds since 1/1/1970.
            var timestamp = (int)((DateTime.UtcNow - epochUtc).TotalSeconds);

            // Add all the OAuth headers we'll need to use when constructing the hash.
            data.Add("oauth_consumer_key", consumerKey);
            data.Add("oauth_signature_method", "HMAC-SHA1");
            data.Add("oauth_timestamp", timestamp.ToString());
            data.Add("oauth_nonce", "a"); // Required, but Twitter doesn't appear to use it, so "a" will do.
            data.Add("oauth_token", accessToken);
            data.Add("oauth_version", "1.0");

            // Generate the OAuth signature and add it to our payload.
            data.Add("oauth_signature", GenerateSignature(fullUrl, data));

            // Build the OAuth HTTP Header from the data.
            string oAuthHeader = GenerateOAuthHeader(data);

            // Build the form data (exclude OAuth stuff that's already in the header).
            var formData = new FormUrlEncodedContent(data.Where(kvp => !kvp.Key.StartsWith("oauth_")));

            return SendRequest(fullUrl, oAuthHeader, formData);
        }

        /// <summary>
        /// Generate an OAuth signature from OAuth header values.
        /// </summary>
        string GenerateSignature(string url, Dictionary<string, string> data)
        {
            var sigString = string.Join(
                "&",
                data
                    .Union(data)
                    .Select(kvp => string.Format("{0}={1}", Uri.EscapeDataString(kvp.Key), Uri.EscapeDataString(kvp.Value)))
                    .OrderBy(s => s)
            );

            var fullSigData = string.Format(
                "{0}&{1}&{2}",
                "POST",
                Uri.EscapeDataString(url),
                Uri.EscapeDataString(sigString.ToString())
            );

            return Convert.ToBase64String(sigHasher.ComputeHash(new ASCIIEncoding().GetBytes(fullSigData.ToString())));
        }

        /// <summary>
        /// Generate the raw OAuth HTML header from the values (including signature).
        /// </summary>
        string GenerateOAuthHeader(Dictionary<string, string> data)
        {
            return "OAuth " + string.Join(
                ", ",
                data
                    .Where(kvp => kvp.Key.StartsWith("oauth_"))
                    .Select(kvp => string.Format("{0}=\"{1}\"", Uri.EscapeDataString(kvp.Key), Uri.EscapeDataString(kvp.Value)))
                    .OrderBy(s => s)
            );
        }

        /// <summary>
        /// Send HTTP Request and return the response.
        /// </summary>
        async Task<string> SendRequest(string fullUrl, string oAuthHeader, FormUrlEncodedContent formData)
        {
            using (var http = new HttpClient())
            {
                http.DefaultRequestHeaders.Add("Authorization", oAuthHeader);

                var httpResp = await http.PostAsync(fullUrl, formData);
                var respBody = await httpResp.Content.ReadAsStringAsync();

                return respBody;
            }
        }
    }
}
